-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AloftGuildText")
local SML = AceLibrary("SharedMedia-1.0")
local SML_font = SML:List("font")

-----------------------------------------------------------------------------

L:RegisterTranslations("enUS", function() return {
	["Guild Text"] = true,
	["Guild text options"] = true,
	["Enable"] = true,
	["Show guild information on nameplates"] = true,
	
	["Attach to Name"] = true,
	["Attaches guild text to name"] = true,
	["Attach Format"] = true,
	["Selects the text format to use"] = true,
	["Name - Guild"] = true,
	["Name (Guild)"]  = true,

	["Short"] = true,
	["Use short guild names"] = true,
	["Set Short Name"] = true,
	["<Any String>"] = true,
	["Sets the short name for your current target"] = true,
	["Show Own"] = true,
	["Show the guild names of your own guild members"] = true,

	["Typeface"] = true,
	["Guild text typeface options"] = true,
	["Font"] = true,
	["Sets the font for guild text"] = true,
	["Font Size"] = true,
	["Sets the font height of the guild text"] = true,
	["Font Shadow"] = true,
	["Show font shadow on guild text"] = true,
	["Outline"] = true,
	["Sets the outline for guild text"] = true,

	["Position"] = true,
	["Adjust guild text position"] = true,
	["Anchor"] = true,
	["Sets the anchor for the guild text"] = true,
	["Anchor To"] = true,
	["Sets the relative point on the health bar to anchor the guild text"] = true,
	["X Offset"] = true,
	["X offset of the guild text"] = true,
	["Y Offset"] = true,
	["Y offset of the guild text"] = true,
	["Alpha"] = true,
	["Sets the alpha of the guild text"] = true,
	["Color"] = true,
	["Sets the guild text color"] = true,

	["None"] = true,
	["Normal"] = true,
	["Thick"] = true,
	
	["NoSetShortName"] = "|cffff0000You must have a target to set the short name|r",
	["ShortNameAdded"] = "Added |cff00ff00%s|r -> |cff00ff00%s|r",

	["Advanced"] = true,
	["Options for expert users"] = true,
	["Format"] = true,
	["Guild tag"] = true,
	["<Any tag string>"] = true,
} end)

-----------------------------------------------------------------------------

local attachFormats =
{
	["NONE"]		= { "[name]", "[name]", "[name]", "[name]" },
	["HYPHEN"]		= { "[name][guild:prefix(\" - \"):hideGuild]", "[name][shortGuild:prefix(\" - \"):hideGuild]", "[name][guild:prefix(\" - \")]", "[name][shortGuild:prefix(\" - \")]" },
	["BRACKETS"]	= { "[name] [guild:hideGuild:bracket]", "[name] [shortGuild:hideGuild:bracket]", "[name] [guild:bracket]", "[name] [shortGuild:bracket]" }
}

local formats		= { "[guild:hideGuild]", "[shortGuild:hideGuild]", "[guild]", "[shortGuild]" }

-----------------------------------------------------------------------------

local AloftGuildText = Aloft:NewModule("GuildText", "AceEvent-2.0")
AloftGuildText.db = Aloft:AcquireDBNamespace("guildText")
Aloft:RegisterDefaults("guildText", "profile", {
	enable			= false,
	useShort		= true,
	showOwn			= true,
	attachToName	= true,
	attachFormat	= "HYPHEN",
	alpha			= 1,
	point			= "BOTTOMLEFT",
	relativeToPoint	= "TOPLEFT",
	offsetX			= 0,
	offsetY			= 0,
	font			= "Arial Narrow",
	fontSize		= 9,
	shadow			= false,
	outline			= "OUTLINE",
	format			= "[guild]",
	color			= { 1, 1, 1 }
})

Aloft.Options.args.guildText =
{
	type = 'group',
	name = L["Guild Text"],
	desc = L["Guild text options"],
	disabled = function() return not Aloft:IsActive() end,
	args =
	{
		enable =
		{
			type = 'toggle',
			name = L["Enable"],
			desc = L["Show guild information on nameplates"],
			get = function() return AloftGuildText.db.profile.enable end,
			set = function(v) AloftGuildText.db.profile.enable = v AloftGuildText:UpdateAttachedText() Aloft:DetermineDataSources() AloftGuildText:UpdateAll() end,
			order = 1
		},
		attachToName =
		{
			type = 'toggle',
			name = L["Attach to Name"],
			desc = L["Attaches guild text to name"],
			disabled = function() return not AloftGuildText.db.profile.enable end,
			get = function() return AloftGuildText.db.profile.attachToName end,
			set = function(v) AloftGuildText.db.profile.attachToName = v AloftGuildText:UpdateAttachedText() Aloft:DetermineDataSources() AloftGuildText:UpdateAll() end,
			order = 2
		},
		attachFormat =
		{
			type = 'text',
			name = L["Attach Format"],
			desc = L["Selects the text format to use"],
			disabled = function() return not AloftGuildText.db.profile.enable or not AloftGuildText.db.profile.attachToName end,
			get = function() return AloftGuildText.db.profile.attachFormat end,
			set = function(v) AloftGuildText.db.profile.attachFormat = v AloftGuildText:UpdateAttachedText() Aloft:DetermineDataSources() AloftGuildText:UpdateAll() end,
			validate = { ["HYPHEN"] = L["Name - Guild"], ["BRACKETS"] = L["Name (Guild)"] },
			order = 3
		},
		short =
		{
			type = 'toggle',
			name = L["Short"],
			desc = L["Use short guild names"],
			disabled = function() return not AloftGuildText.db.profile.enable end,
			get = function() return AloftGuildText.db.profile.useShort end,
			set = function(v) AloftGuildText.db.profile.useShort = v AloftGuildText:UpdateAttachedText() Aloft:DetermineDataSources() AloftGuildText:UpdateAll() end,
			order = 4
		},
		setShort =
		{
			type = 'text',
			name = L["Set Short Name"],
			desc = L["Sets the short name for your current target"],
			usage = L["<Any String>"],
			disabled = function() return not AloftGuildText.db.profile.enable or not AloftGuildText.db.profile.useShort end,
			get = function() return "" end,
			set = function(v) 
					if UnitExists("target") then
						local targetGuild = GetGuildInfo("target") or ""
						Aloft:GetModule("ShortGuildData"):AddData(targetGuild, v)
						Aloft:Print(string.format(L["ShortNameAdded"], targetGuild, v))
					else
						Aloft:Print(L["NoSetShortName"])
					end
				  end,
			order = 5
		},
		showOwn =
		{
			type = 'toggle',
			name = L["Show Own"],
			desc = L["Show the guild names of your own guild members"],
			disabled = function() return not AloftGuildText.db.profile.enable end,
			get = function() return AloftGuildText.db.profile.showOwn end,
			set = function(v) AloftGuildText.db.profile.showOwn = v AloftGuildText:UpdateAttachedText() Aloft:DetermineDataSources() AloftGuildText:UpdateAll() end,
			order = 6
		},
		typeface =
		{
			type = "group",
			name = L["Typeface"],
			desc = L["Guild text typeface options"],
			disabled = function() return not AloftGuildText.db.profile.enable or AloftGuildText.db.profile.attachToName end,
			args =
			{
				font =
				{
					type = "text", 
					name = L["Font"], 
					desc = L["Sets the font for guild text"],
					get = function() return AloftGuildText.db.profile.font end,
					set = function(v) AloftGuildText.db.profile.font = v AloftGuildText:UpdateAll() end,
					validate = SML_font,
				},
				fontSize =
				{
					type = 'range',
					name = L["Font Size"],
					desc = L["Sets the font height of the guild text"],
					max = 16,
					min = 5,
					step = 1,
					get = function() return AloftGuildText.db.profile.fontSize end,
					set = function(value) AloftGuildText.db.profile.fontSize = value AloftGuildText:UpdateAll() end
				},
				shadow =
				{
					type = 'toggle',
					name = L["Font Shadow"],
					desc = L["Show font shadow on guild text"],
					get = function() return AloftGuildText.db.profile.shadow end,
					set = function(v) AloftGuildText.db.profile.shadow = v AloftGuildText:UpdateAll() end
				},
				outline =
				{
					type = 'text',
					name = L["Outline"],
					desc = L["Sets the outline for guild text"],
					get = function() return AloftGuildText.db.profile.outline end,
					set = function(value) AloftGuildText.db.profile.outline = value AloftGuildText:UpdateAll() end,
					validate = { [""] = L["None"], ["OUTLINE"] = L["Normal"], ["THICKOUTLINE"] = L["Thick"] },
				},
			},
		},
		position =
		{
			type = 'group',
			name = L["Position"],
			desc = L["Adjust guild text position"],
			disabled = function() return not AloftGuildText.db.profile.enable or AloftGuildText.db.profile.attachToName end,
			args =
			{
				anchor = {
					type = "text", 
					name = L["Anchor"], 
					desc = L["Sets the anchor for the guild text"],
					get = function() return AloftGuildText.db.profile.point end,
					set = function(v) AloftGuildText.db.profile.point = v AloftGuildText:UpdateAll() end,
					validate = {"TOPLEFT", "TOP", "TOPRIGHT", "LEFT", "CENTER", "RIGHT", "BOTTOMLEFT", "BOTTOM", "BOTTOMRIGHT"},
				},
				anchorto = {
					type = "text", 
					name = L["Anchor To"], 
					desc = L["Sets the relative point on the health bar to anchor the guild text"],
					get = function() return AloftGuildText.db.profile.relativeToPoint end,
					set = function(v) AloftGuildText.db.profile.relativeToPoint = v AloftGuildText:UpdateAll() end,
					validate = {"TOPLEFT", "TOP", "TOPRIGHT", "LEFT", "CENTER", "RIGHT", "BOTTOMLEFT", "BOTTOM", "BOTTOMRIGHT"},
				},
				offsetX =
				{
					type = 'range',
					name = L["X Offset"],
					desc = L["X offset of the guild text"],
					min = -32,
					max = 32,
					step = 1,
					get = function() return AloftGuildText.db.profile.offsetX end,
					set = function(v) AloftGuildText.db.profile.offsetX = v AloftGuildText:UpdateAll() end
				},
				offsetY =
				{
					type = 'range',
					name = L["Y Offset"],
					desc = L["Y offset of the guild text"],
					min = -32,
					max = 32,
					step = 1,
					get = function() return AloftGuildText.db.profile.offsetY end,
					set = function(v) AloftGuildText.db.profile.offsetY = v AloftGuildText:UpdateAll() end
				},
			},
		},
		alpha =
		{
			type = 'range',
			name = L["Alpha"],
			desc = L["Sets the alpha of the guild text"],
			min = 0,
			max = 1.0,
			step = 0.05,
			disabled = function() return not AloftGuildText.db.profile.enable or AloftGuildText.db.profile.attachToName end,
			get = function() return AloftGuildText.db.profile.alpha end,
			set = function(v) AloftGuildText.db.profile.alpha = v AloftGuildText:UpdateAll() end
		},
		color =
		{
			type = 'color',
			name = L["Color"],
			desc = L["Sets the guild text color"],
			disabled = function() return not AloftGuildText.db.profile.enable or AloftGuildText.db.profile.attachToName end,
			get = function() return unpack(AloftGuildText.db.profile.color) end,
			set = function(r, g, b, a) AloftGuildText.db.profile.color = { r, g, b, a } AloftGuildText:UpdateAll() end,
		},
		advanced  =
		{
			type = 'group',
			name = L["Advanced"],
			desc = L["Options for expert users"],
			order = -1,
			disabled = function() return not AloftGuildText.db.profile.enable or AloftGuildText.db.profile.attachToName end,
			args =
			{
				format = 
				{
					type = 'text',
					name = L["Format"],
					desc = L["Guild tag"],
					usage = L["<Any tag string>"],
					get = function() return AloftGuildText.db.profile.format end,
					set = function(v) AloftGuildText.db.profile.format = v Aloft:DetermineDataSources() AloftGuildText:UpdateAll() end,
				}
			},
		},
	},
}

-----------------------------------------------------------------------------

local textMethod = nil
local textMethodData = { }
local textRegionPool = { }

-----------------------------------------------------------------------------

function AloftGuildText:UpdateAll()
	for aloftData in Aloft:IterateNameplates() do
		if self.db.profile.enable then
			if self.db.profile.attachToName then
				self:OnNameplateHide(aloftData)
			else
				if aloftData.guildTextRegion then
					self:PrepareText(aloftData.guildTextRegion, self.db.profile)
					self:PlaceFrame(aloftData.guildTextRegion, aloftData.healthBar, self.db.profile)
				end
				if aloftData:IsShown() then
					self:UpdateText(aloftData)
				end
			end
		else
			self:OnNameplateHide(aloftData)
		end
	end	
    for textRegion in pairs(textRegionPool) do
        self:PrepareText(textRegion, self.db.profile)
    end
	Aloft:GetModule("NameText"):UpdateAll()
end

function AloftGuildText:UpdateAttachedText()
	local nameTextModule = Aloft:GetModule("NameText")
	
	local attachFormatIndex = 1
	if self.db.profile.useShort then attachFormatIndex = 2 end
	if self.db.profile.showOwn then attachFormatIndex = attachFormatIndex+2 end

	local attachFormat = self.db.profile.attachFormat
	if not self.db.profile.enable or not self.db.profile.attachToName then
		attachFormat = "NONE"
	end
	
	nameTextModule.db.profile.format = attachFormats[attachFormat][attachFormatIndex]
	self.db.profile.format = formats[attachFormatIndex]
end

-----------------------------------------------------------------------------

function AloftGuildText:RegisterEvents()
	self:UnregisterAllEvents()
	
	self:RegisterEvent("Aloft:SetAll", function(type, value)
										if self.db.profile[type] then
											self.db.profile[type] = value
											self:UpdateAll()
										end
									   end)

	if self.db.profile.enable then
		if not self.db.profile.attachToName then
			self:RegisterEvent("Aloft:OnNameplateShow", "UpdateText")
			self:RegisterEvent("Aloft:OnNameplateHide", "OnNameplateHide")
			for events in pairs(textMethodData.events) do
				self:RegisterEvent(events, "UpdateText")
			end
		end
	end
end

function AloftGuildText:OnEnable()
	self:UpdateAll()
end

function AloftGuildText:OnDisable()
	self:UnregisterAllEvents()
	
	for aloftData in Aloft:IterateNameplates() do
		local guildTextRegion = aloftData.guildTextRegion
		if guildTextRegion then
			guildTextRegion:Hide()
		end
	end
end

local dataRequiredList = { }
function AloftGuildText:RequiresData()
	local profile = self.db.profile

	textMethodData = AloftTag:CreateTag(profile.format)
	textMethod = textMethodData.method

	self:RegisterEvents()

	if not profile.enable then
		return
	end

	for i=1,#dataRequiredList do
		dataRequiredList[i] = nil
	end

	for data in pairs(textMethodData.data) do
		table.insert(dataRequiredList, data)
	end

	return unpack(dataRequiredList)
end

-----------------------------------------------------------------------------

function AloftGuildText:UpdateText(aloftData)
    local text = textMethod(aloftData)

    if text ~= "" then 
        -- Check if this nameplate already has a text region assigned to it
        local guildTextRegion = aloftData.guildTextRegion
        if not guildTextRegion then
            guildTextRegion = next(textRegionPool)
            if guildTextRegion then
                -- We have a recycled text region. Text format already setup, just need to place it
                textRegionPool[guildTextRegion] = nil
                guildTextRegion:SetParent(aloftData.nameplateFrame)
                guildTextRegion:Show()
            else
                -- We need a new text region.
                guildTextRegion = aloftData:CreateFontString()
                self:PrepareText(guildTextRegion, self.db.profile)
            end
            self:PlaceFrame(guildTextRegion, aloftData.healthBar, self.db.profile)
            aloftData.guildTextRegion = guildTextRegion
        end
        guildTextRegion:SetText(text)
    else
        self:OnNameplateHide(aloftData)
    end
end 

function AloftGuildText:OnNameplateHide(aloftData)
    local guildTextRegion = aloftData.guildTextRegion
    if guildTextRegion then
        guildTextRegion:Hide()
        textRegionPool[guildTextRegion] = true
        aloftData.guildTextRegion = nil
    end
end

-----------------------------------------------------------------------------
